"""logger-color
Different logging levels for stram (terminal) and file. Color the log messages.
Has additional levels for diagnostics: DIAG_INFO, DIAG_WARNING.

============== ===== ===========
Level		   Int   Color
============== ===== ===========
DEBUG          10    gray
INFO           20    green
WARNING		   30    yellow
ERROR		   40    red
CRITICAL	   50    bold red
DIAG_INFO      21    blue
DIAG_WARNING   31    blue bold
============== ===== ===========
"""

from logger_color.clogger import CLogger
from logger_color.functions import debug, info, warning, error, critical
from logger_color.functions import start_logging, diag_info, diag_warning

__all__ = [
    "CLogger",
    "critical",
    "debug",
    "diag_info",
    "diag_warning",
    "error",
    "info",
    "start_logging",
    "warning",
]

__title__ = "logger_color"
__version__ = "0.0.6"
__date__ = "2022-10-28"

__summary__ = "Different logging levels for stram (terminal) and file. Color the log messages. " \
              "Has additional levels for diagnostics: DIAG_INFO, DIAG_WARNING."
__author__ = "Vladimir Prusakov"
__email__ = "vladimir.prusakovs@gmail.com"
__url__ = "https://github.com/vladimirs-git/logger"
__download_url__ = f"{__url__}/archive/refs/tags/{__version__}.tar.gz"
__license__ = "MIT"



"""Colored logging formatter"""

import re
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL, NOTSET
from logging import Formatter, LogRecord
from typing import Dict, Union

DiStr = Dict[int, str]
DiFormatter = Dict[int, Formatter]

NO_COLOR = "\33[0m"
GRAY = "\33[90m"
BLUE = "\33[34m"
BLUE_BOLD = "\33[34;1m"
GREEN = "\33[1;32m"
RED = "\33[31m"
RED_BOLD = "\33[31;1m"
YELLOW = "\33[33m"

DIAG_INFO = 21
DIAG_WARNING = 31

COLORS = {
    DEBUG: GRAY,
    INFO: GREEN,
    WARNING: YELLOW,
    ERROR: RED,
    CRITICAL: RED_BOLD,

    NOTSET: NO_COLOR,
    DIAG_INFO: BLUE,
    DIAG_WARNING: BLUE_BOLD,
}
COLORS_ = {
    "DEBUG": GRAY,
    "INFO": GREEN,
    "WARNING": YELLOW,
    "ERROR": RED,
    "CRITICAL": RED_BOLD,

    "NOTSET": NO_COLOR,
    "DIAG_INFO": BLUE,
    "DIAG_WARNING": BLUE_BOLD,
}


class CFormatter(Formatter):
    """Colored logging formatter"""

    def __init__(self, fmt: str):
        super().__init__()
        self._colors: DiStr = COLORS.copy()
        self.formatters: DiFormatter = self._init_formatters(fmt=fmt)

    # ============================= init =============================

    def _init_formatters(self, fmt: str) -> DiFormatter:
        """Init colored logging formatters by logging levels"""
        date_fmt = "%Y-%m-%d %H:%M:%S"
        formatters: DiFormatter = {}  # return
        for level, color in self._colors.items():
            # no logging level
            if level == NOTSET:
                formatters[NOTSET] = Formatter(fmt=fmt, datefmt=date_fmt)
                continue
            # logging level in [DEBUG, ..., CRITICAL]. Change color
            fmt_ = fmt
            for fld in ["levelname", "levelno"]:
                if fld in fmt_:
                    regex = f"(%\\({fld}\\).*?s)"
                    fmt_ = re.sub(regex, f"{color}\\1{NO_COLOR}", fmt_)
            formatter = Formatter(fmt=fmt_, datefmt=date_fmt)
            formatters[level] = formatter
        return formatters

    # =========================== methods ============================

    def format(self, record: LogRecord) -> str:
        """Change color of logging record"""
        orig_msg: str = record.msg
        orig_args = record.args
        self._update_record(record)
        formatter: Formatter = self._get_formatter(level=record.levelno)
        msg: str = formatter.format(record)
        record.msg = orig_msg
        record.args = orig_args
        return msg

    # =========================== helpers ============================

    def _get_formatter(self, level: int) -> Formatter:
        """Returns colored formatter by logging level"""
        formatter = self.formatters.get(level) or self.formatters[NOTSET]
        return formatter

    @staticmethod
    def _update_record(record: LogRecord) -> None:
        """Updates data in LogRecord. Change message color.
        if record.levelno == 21 > record.levelname = DIAG_INFO
        if record.levelno == 31 > record.levelname = DIAG_WARNING
        """
        if record.levelno not in [DIAG_INFO, DIAG_WARNING]:
            return
        record.msg = f"{BLUE}{record.msg}{NO_COLOR}"
        if record.levelno == DIAG_INFO:
            record.levelname = "DIAG_I"
        elif record.levelno == DIAG_WARNING:
            record.levelname = "DIAG_W"


UFormatter = Union[CFormatter, Formatter]



"""Logging functions"""

import logging

from logger_color.cformatter import DIAG_INFO, DIAG_WARNING
from logger_color.clogger import CLogger


# noinspection PyIncorrectDocstring
def start_logging(**kwargs):
    """Start logging with the specified parameters
    :param filename: Writes messages to this file. By default, logging only to the terminal
    :type filename: str

    :param mode: "w" - Writes to file (default)
                 "a" - Appends to file
    :type mode: str

    :param level: Logging level for the terminal: "DEBUG", "INFO" (default), "WARNING", "ERROR",
        "CRITICAL", "DIAG_INFO", "DIAG_WARNING"
    :type level: str

    :param level_file: Logging level for the file. By default, the same as `level` parameter
    :type level_file: str

    :param color: True  - Prints a colored message to the terminal (default)
                  False - Prints a monochrome message to the terminal

    :param debug: True  - Sets logging level to DEBUG (rewrites `level` parameter),
                          works the same as level="DEBUG"
                  False - Gets logging level from `level` parameter (default)
    :type debug: bool
    """
    CLogger(**kwargs).start()


def debug(*args, **kwargs) -> None:
    """Log a message with severity "DEBUG" """
    logging.debug(*args, **kwargs)


def diag_info(msg, *args, **kwargs) -> None:
    """Log diagnostic message with severity "INFO"
    :param msg: Logging message
    """
    # noinspection PyProtectedMember
    logging.root._log(DIAG_INFO, msg, args, **kwargs)


def diag_warning(msg, *args, **kwargs) -> None:
    """Log diagnostic message with severity "WARNING"
    :param msg: Logging message
    """
    # noinspection PyProtectedMember
    logging.root._log(DIAG_WARNING, msg, args, **kwargs)


def info(*args, **kwargs) -> None:
    """Log a message with severity "INFO" """
    logging.info(*args, **kwargs)


def warning(*args, **kwargs) -> None:
    """Log a message with severity "WARNING" """
    logging.warning(*args, **kwargs)


def error(*args, **kwargs) -> None:
    """Log a message with severity "ERROR" """
    logging.error(*args, **kwargs)


def critical(*args, **kwargs) -> None:
    """Log a message with severity "CRITICAL" """
    logging.critical(*args, **kwargs)



"""Package setup"""

import pathlib

from setuptools import setup, find_packages  # type: ignore

import logger_color as package

VERSION = "0.0.6"
PACKAGE = package.__title__
PACKAGE_ = package.__title__.lower().replace("-", "_")  # PEP 503 normalization
ROOT = pathlib.Path(__file__).parent.resolve()
README = "README.rst"

if __name__ == "__main__":
    setup(
        name=PACKAGE_,
        packages=[PACKAGE_],
        package_data={PACKAGE_: ["py.typed"]},
        version=VERSION,
        description=package.__summary__,
        license=package.__license__,
        long_description=open(README).read(),
        long_description_content_type="text/x-rst",
        author=package.__author__,
        author_email=package.__email__,
        url=package.__url__,
        download_url=package.__download_url__,
        keywords="logging",
        python_requires=">=3.8",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Natural Language :: English",
        ],
    )



"""Color Logger"""

import logging
import os
import sys
from logging import DEBUG, INFO
from logging import Formatter
from pathlib import Path

from logger_color.cformatter import CFormatter, UFormatter


class CLogger:
    """Color Logger"""

    def __init__(self, **kwargs):
        """Color Logger
        :param filename: Writes messages to this file. By default, logging only to the terminal
        :type filename: str

        :param mode: "w" - Writes to file (default)
                     "a" - Appends to file
        :type mode: str

        :param level: Logging level for the terminal: "DEBUG", "INFO" (default), "WARNING", "ERROR",
            "CRITICAL", "DIAG_INFO", "DIAG_WARNING"
        :type level: str

        :param level_file: Logging level for the file. By default, the same as `level` parameter
        :type level_file: str

        :param color: True  - Prints a colored message to the terminal (default)
                      False - Prints a monochrome message to the terminal

        :param debug: True  - Sets logging level to DEBUG (rewrites `level` parameter),
                              works the same as level="DEBUG"
                      False - Gets logging level from `level` parameter (default)
        :type debug: bool
        """
        self.filename: str = kwargs.get("filename") or ""
        self.level: int = self._init_level(**kwargs)
        self.level_file: int = self._init_level_file(**kwargs)
        self.mode: str = self._init_mode(**kwargs)
        self.color: bool = self._init_color(**kwargs)

    # ============================= init =============================

    @staticmethod
    def _init_color(**kwargs) -> bool:
        """Init logging color"""
        color = kwargs.get("color")
        if color is None:
            color = True
        return bool(color)

    def _init_formatter(self, level: int, color: bool = False) -> UFormatter:
        """Init logging formatter"""
        if level <= DEBUG:
            format_ = "%(asctime)s.%(msecs)03d %(levelname)-8s %(module)s.%(funcName)s: %(message)s"
        else:
            format_ = "%(asctime)s %(levelname)-8s %(message)s"
        if color and self.color:
            formatter: UFormatter = CFormatter(fmt=format_)
        else:
            formatter = Formatter(fmt=format_, datefmt="%Y-%m-%d %H:%M:%S")
        return formatter

    @staticmethod
    def _init_level(**kwargs) -> int:
        """Init logging level, by default INFO"""
        debug = kwargs.get("debug")
        if debug is None:
            debug = kwargs.get("DEBUG")
        if bool(debug):
            return DEBUG
        level = kwargs.get("level")
        if level is None:
            return INFO
        if isinstance(level, int):
            return level
        if isinstance(level, str):
            level = level.upper()
        # noinspection PyUnresolvedReferences,PyProtectedMember
        level_ = logging._nameToLevel.get(level)
        if isinstance(level_, int):
            return level_
        return INFO

    def _init_level_file(self, **kwargs) -> int:
        """Init logging level for file, by default the same as level for the terminal"""
        level = kwargs.get("level_file")
        if level is None:
            return self.level
        if isinstance(level, int):
            return level
        if isinstance(level, str):
            level = level.upper()
        # noinspection PyUnresolvedReferences,PyProtectedMember
        level_ = logging._nameToLevel.get(level)
        if isinstance(level_, int):
            return level_
        return self.level

    @staticmethod
    def _init_mode(**kwargs) -> str:
        """Init logging mode"""
        mode = kwargs.get("mode")
        if mode is None:
            mode = "w"
        expected = ["a", "w"]
        if mode not in expected:
            raise ValueError(f"{mode=}, {expected=}")
        return mode

    # =========================== methods ============================

    def start(self) -> None:
        """Starts logging"""
        logger = logging.getLogger()
        min_level = min(self.level, self.level_file)
        logger.setLevel(level=min_level)
        self._remove_handlers()
        self._add_stream_handler()
        self._add_file_handler()

    # =========================== helpers ============================

    @staticmethod
    def _remove_handlers() -> None:
        """Removes all logging handlers"""
        logger = logging.getLogger()
        while logger.handlers:
            handler = logger.handlers[0]
            logger.removeHandler(handler)

    @staticmethod
    def _check_dir_write(path: str) -> None:
        """Check directory write permission"""
        if not os.path.exists(path):
            raise OSError(f"absent {path=}")
        if not os.access(path, os.W_OK):
            raise OSError(f"write permission error {path=}")

    def _add_stream_handler(self) -> None:
        """Adds StreamHandler to Logger"""
        logger = logging.getLogger()
        formatter = self._init_formatter(level=self.level, color=self.color)
        handler = logging.StreamHandler(stream=sys.stdout)
        handler.setLevel(self.level)
        handler.setFormatter(fmt=formatter)
        logger.addHandler(handler)

    def _add_file_handler(self) -> None:
        """Add FileHandler to Logger, logging WARNING if it does not have write-permissions"""
        if not self.filename:
            return
        path = str(Path(self.filename).parent)
        logger = logging.getLogger()
        try:
            self._check_dir_write(path)
            handler = logging.FileHandler(filename=self.filename, mode=self.mode)
            formatter = self._init_formatter(level=self.level_file)
            handler.setFormatter(fmt=formatter)
            handler.setLevel(self.level_file)
            logger.addHandler(handler)
        except OSError as ex:
            logging.warning(str(ex))
